"""
Based heavily on Chemspipy by Matt Swain
https://github.com/mcs07/ChemSpiPy
"""
from pura.services import Service
from pura.compound import CompoundIdentifier, CompoundIdentifierType
from aiohttp import ClientSession
import asyncio
from typing import Optional, List
import os
import logging
from dotenv import load_dotenv

load_dotenv()
logger = logging.getLogger(__name__)

#: 2D coordinate dimensions
MOL2D = "2d"
#: 3D coordinate dimensions
MOL3D = "3d"
#: Both coordinate dimensions
BOTH = "both"

#: Ascending sort direction
ASCENDING = "ascending"
#: Descending sort direction
DESCENDING = "descending"

#: Record ID sort order
RECORD_ID = "record_id"
#: CSID sort order (same as RECORD_ID, kept for backwards compatibility)
CSID = "csid"
#: Mass defect sort order
MASS_DEFECT = "mass_defect"
#: Molecular weight sort order
MOLECULAR_WEIGHT = "molecular_weight"
#: Reference count sort order
REFERENCE_COUNT = "reference_count"
#: Datasource count sort order
DATASOURCE_COUNT = "datasource_count"
#: Pubmed count sort order
PUBMED_COUNT = "pubmed_count"
#: RSC count sort order
RSC_COUNT = "rsc_count"

#: Map sort directions to strings required by REST API.
DIRECTIONS = {ASCENDING: "ascending", DESCENDING: "descending"}

#: Map sort orders to strings required by REST API.
ORDERS = {
    RECORD_ID: "recordId",
    CSID: "recordId",
    MASS_DEFECT: "massDefect",
    MOLECULAR_WEIGHT: "molecularWeight",
    REFERENCE_COUNT: "referenceCount",
    DATASOURCE_COUNT: "dataSourceCount",
    PUBMED_COUNT: "pubMedCount",
    RSC_COUNT: "rscCount",
}

#: All available compound details fields.
FIELDS = [
    "SMILES",
    "Formula",
    "AverageMass",
    "MolecularWeight",
    "MonoisotopicMass",
    "NominalMass",
    "CommonName",
    "ReferenceCount",
    "DataSourceCount",
    "PubMedCount",
    "RSCCount",
    "Mol2D",
    "Mol3D",
]
IDENTIFIER_TYPES = [
    CompoundIdentifierType.CHEMSPIDER_ID,
    CompoundIdentifierType.SMILES,
    CompoundIdentifierType.INCHI,
    CompoundIdentifierType.INCHI_KEY,
]


class ChemSpider(Service):
    api_url = "https://api.rsc.org"
    api_version = "v1"

    """

    Arguments
    ---------
    token : str, optional
        Chemspider API token. If not passed, will try to read from the
        environmental variable `CHEMSPIDER_TOKEN`.
    """

    def __init__(self, token: Optional[str] = None) -> None:
        if token is None:
            token = os.environ.get("CHEMSPIDER_TOKEN")
            if token is None:
                raise ValueError(
                    "No Chemspider API token passed or found in the environment."
                )
        self.token = token

    async def resolve_compound(
        self,
        session: ClientSession,
        input_identifier: CompoundIdentifier,
        output_identifier_types: List[CompoundIdentifierType],
    ) -> List[CompoundIdentifierType]:
        if not any(
            [
                output_identifier_type in IDENTIFIER_TYPES
                for output_identifier_type in output_identifier_types
            ]
        ):
            raise ValueError(
                f"{output_identifier_types} are invalid output identifier type(s) for ChemSpider"
            )

        query_id = await self.filter_name(
            session, input_identifier.value, order=None, direction=ASCENDING
        )
        record_ids = await self.filter_results(session, query_id=query_id)

        resolved_identifiers = []
        if CompoundIdentifierType.CHEMSPIDER_ID in output_identifier_types:
            resolved_identifiers += [
                CompoundIdentifier(
                    identifier_type=CompoundIdentifierType.CHEMSPIDER_ID,
                    value=record_id,
                )
                for record_id in record_ids
            ]
        for record_id in record_ids:
            details = await self.get_details(session, record_id)
            if CompoundIdentifierType.SMILES in output_identifier_types:
                resolved_identifiers += [
                    CompoundIdentifier(
                        identifier_type=CompoundIdentifierType.SMILES,
                        value=details.get("smiles"),
                    )
                ]
            if CompoundIdentifierType.INCHI in output_identifier_types:
                mol_2d = details["mol2D"]
                value = await self.convert(session, mol_2d, "Mol", "InChI")
                resolved_identifiers += [
                    CompoundIdentifier(
                        value=value, identifier_type=CompoundIdentifierType.INCHI
                    )
                ]
            if CompoundIdentifierType.INCHI_KEY in output_identifier_types:
                mol_2d = details["mol2D"]
                value = await self.convert(session, mol_2d, "Mol", "InChIKey")
                resolved_identifiers += [
                    CompoundIdentifier(
                        identifier_type=CompoundIdentifierType.INCHI_KEY, value=value
                    )
                ]

        return resolved_identifiers

    async def request(
        self,
        session: ClientSession,
        method: str,
        api: str,
        namespace: str,
        endpoint: str,
        params: dict = None,
        json: dict = None,
    ):

        # if params is None:
        #     params = {}
        # if json is None:
        #     json = {}

        # Construct request URL
        url = "{}/{}/{}/{}/{}".format(
            self.api_url, api, self.api_version, namespace, endpoint
        )

        # Set apikey header
        headers = {
            "apikey": self.token,
        }

        logger.debug("{} : {} : {} : {}".format(url, headers, params, json))

        # Make request
        async with session.request(
            method=method, url=url, json=json, params=params, headers=headers
        ) as resp:
            logger.debug(f"Response status: {resp.status}")
            response = await resp.json()

        return response

    async def filter_name(
        self, session: ClientSession, input_identifier, order=None, direction=None
    ):
        json = {
            "name": input_identifier,
            "orderBy": ORDERS.get(order),
            "orderDirection": DIRECTIONS.get(direction),
        }
        response = await self.request(
            session,
            method="post",
            api="compounds",
            namespace="filter",
            endpoint="name",
            json=json,
        )
        return response["queryId"]

    async def filter_results(self, session: ClientSession, query_id):
        """Get filter results using a query ID that was returned by a previous filter request.

        :param string query_id: Query ID from a previous filter request.
        :param int start: Zero-based results offset.
        :param int count: Number of results to return.
        :return: List of results.
        :rtype: list[int]
        """
        endpoint = "{}/results".format(query_id)
        response = await self.request(
            session,
            method="get",
            api="compounds",
            namespace="filter",
            endpoint=endpoint,
        )
        return response["results"]

    async def convert(self, session: ClientSession, input, input_format, output_format):
        """Convert a chemical from one format to another.
        Format: ``SMILES``, ``InChI``, ``InChIKey`` or ``Mol``.
        Allowed conversions: from InChI to InChIKey, from InChI to Mol file, from InChI to SMILES, from InChIKey to
        InChI, from InChIKey to Mol file, from Mol file to InChI, from Mol file to InChIKey, from SMILES to InChI.
        :param string input: Input chemical.
        :param string input_format: Input format.
        :param string output_format: Output format.
        :return: Input chemical in output format.
        :rtype: string
        """
        json = {
            "input": input,
            "inputFormat": input_format,
            "outputFormat": output_format,
        }
        response = await self.request(
            session,
            method="post",
            api="compounds",
            namespace="tools",
            endpoint="convert",
            json=json,
        )
        return response["output"]

    async def get_details(self, session: ClientSession, record_id, fields=FIELDS):
        """Get details for a compound record.
        The available fields are listed in :data:`~chemspipy.api.FIELDS`.
        :param int record_id: Record ID.
        :param list[string] fields: (Optional) List of fields to include in the result.
        :return: Record details.
        :rtype: dict
        """
        params = {"fields": ",".join(fields)}
        endpoint = "{}/details".format(record_id)
        response = await self.request(
            session,
            method="get",
            api="compounds",
            namespace="records",
            endpoint=endpoint,
            params=params,
        )
        return response



from pura.units import Mass, Amount, Volume
from rdkit import Chem
from pydantic import BaseModel
from typing import Optional, List, Any
from enum import Enum
import re
import warnings


class CompoundIdentifierType(Enum):
    UNSPECIFIED = 0
    CUSTOM = 1
    # Simplified molecular-input line-entry system
    SMILES = 2
    # IUPAC International Chemical Identifier
    INCHI = 3
    # Molblock from a MDL Molfile V3000
    MOLBLOCK = 4
    # Chemical name following IUPAC nomenclature recommendations
    IUPAC_NAME = 5
    # Any accepted common name, trade name, etc.
    NAME = 6
    # Chemical Abstracts Service Registry Number (with hyphens)
    CAS_NUMBER = 7
    # PubChem Compound ID number
    PUBCHEM_CID = 8
    # ChemSpider ID number
    CHEMSPIDER_ID = 9
    # ChemAxon extended SMILES
    CXSMILES = 10
    # IUPAC International Chemical Identifier key
    INCHI_KEY = 11
    # XYZ molecule file
    XYZ = 12
    # UniProt ID (for enzymes)
    UNIPROT_ID = 13
    # Protein data bank ID (for enzymes)
    PDB_ID = 14
    # Amino acid sequence (for enzymes).
    AMINO_ACID_SEQUENCE = 15
    # HELM; https:#www.pistoiaalliance.org/helm-notation/.
    HELM = 16
    # SMILES arbitrary target specification
    SMARTS = 17


class Data:
    pass


class Analysis:
    pass


class Source:
    pass


class CompoundIdentifier(BaseModel):
    identifier_type: CompoundIdentifierType
    value: str
    details: Optional[str] = None

    def __eq__(self, other: Any) -> bool:
        return (
            other.identifier_type == self.identifier_type and self.value == other.value
        )


class Compound(BaseModel):
    identifiers: List[CompoundIdentifier]
    # amount: Union[Amount, Mass, Volume]
    amount: Amount = None
    mass: Mass = None
    volume: Volume = None
    # source: Source = None
    # data: Dict[str, Data] = None
    # analyses: Dict[str, Analysis] = None


def unique_identifiers(identifiers: List[CompoundIdentifier]):
    unique_identifiers = []
    for identifier in identifiers:
        if identifier not in unique_identifiers:
            unique_identifiers.append(identifier)
    return unique_identifiers


def standardize_identifier(identifier: CompoundIdentifier):
    if identifier.identifier_type == CompoundIdentifierType.SMILES:
        smi = identifier.value
        # check smi and raise warnings
        if "*" in smi:
            # SMILES example: '*', '*CCC'
            warnings.warn("Warning: * in SMILES string.")
        if "." in smi:
            smi_segs = smi.split(".")
            smi_segs = [re.sub("-\w", "", _) for _ in smi_segs]
            if all([(("+" not in _) & ("-" not in _)) for _ in smi_segs]):
                # SMILES example: '[HH].[HH].[HH].[HH].[HH].[Ir].[MgH2].[MgH2]'
                warnings.warn(
                    "Warning: SMILES of a mixture, rather than a pure compound, was found."
                )
        if ("+" in smi) | ("-" in smi):
            # calculate charge of the compound based on its SMILES
            p = re.findall("\+(.*?)\]", smi)
            n = re.findall("\-(.*?)\]", smi)
            p_total = []
            n_total = []
            for _ in p:
                if str.isdigit(_):
                    p_total.append(int(_))
                elif _ == "":
                    p_total.append(1)
                else:
                    continue
            for _ in n:
                if str.isdigit(_):
                    n_total.append(int(_))
                elif _ == "":
                    n_total.append(1)
                else:
                    continue
            p_total = sum(p_total)
            n_total = -sum(n_total)
            comp_charge = p_total + n_total
            if comp_charge != 0:
                # SMILES example: 'O=C([O-])[O-].[K+]', '[NH4+]', '[H]O[H].[O-][I+2]([O-])[O-]'
                warnings.warn(
                    "Warning: Compound is not electrically neutral based on its SMILES string."
                )
                # NOTE: did not add the function to neutralize charged SMILES, the balance function could do sth like this:
                # ['[Cs+].[H-].[PdH4-2]', 'O=[V+]([O-])O.[NH4+]', '[H]C(=O)[O-].[H]O[H].[Rh+2]', '[O-][I+2]([O-])[O-]'] ->
                # ['[Cs+].[Cs+].[Cs+].[H-].[PdH4-2]', 'O=[V+]([O-])O.[NH4+].[OH-]', '[H]C(=O)[O-].[H]C(=O)[O-].[H]O[H].[Rh+2]', '[O-][I+2]([O-])[O-].[H+]']
                # if this feature is necessary we can add later
        mol = Chem.MolFromSmiles(smi)
        if mol is not None:
            for a in mol.GetAtoms():
                if a.GetIsotope() != 0:
                    # SMILES example: '[2H]P([2H])[2H]'
                    warnings.warn("Warning: SMILES string contains isotopes.")
            Chem.SanitizeMol(mol)
            mol.UpdatePropertyCache(strict=False)
            identifier.value = Chem.MolToSmiles(mol)



from pura.services import Service
from pura.compound import CompoundIdentifier, CompoundIdentifierType
from pura.utils import inverse_map
from aiohttp import ClientSession
from typing import List, Union
from aiohttp.web_exceptions import HTTPNotFound
import logging

BASE_API = "https://rboq1qukh0.execute-api.us-east-2.amazonaws.com"

logger = logging.getLogger(__name__)

INPUT_IDENTIFIER_MAP = {
    CompoundIdentifierType.CAS_NUMBER: "cas_rn",
    # CompoundIdentifierType.SMILES: "smiles",
    # CompoundIdentifierType.INCHI: "stdinchi",
    # CompoundIdentifierType.INCHI_KEY: "stdinchikey",
}

OUTPUT_IDENTIFIER_MAP = {
    CompoundIdentifierType.CAS_NUMBER: "rn",
    CompoundIdentifierType.SMILES: "canonicalSmile",
    CompoundIdentifierType.INCHI: "stdinchi",
    CompoundIdentifierType.INCHI_KEY: "inchiKey",
}


class CAS(Service):
    async def resolve_compound(
        self,
        session: ClientSession,
        input_identifier: CompoundIdentifier,
        output_identifier_types: List[CompoundIdentifierType],
    ) -> List[Union[CompoundIdentifier, None]]:
        input_identifier_cas = INPUT_IDENTIFIER_MAP.get(
            input_identifier.identifier_type
        )
        if input_identifier_cas is None:
            raise ValueError(
                f"{input_identifier.identifier_type} is not one of the valid identifier types for the CAS."
            )

        output_representations = [
            OUTPUT_IDENTIFIER_MAP.get(output_identifier_type)
            for output_identifier_type in output_identifier_types
        ]
        if not any(output_representations):
            raise ValueError(
                f"{output_identifier_types} are not one of the valid identifier types for CAS."
            )

        results = await common_chemistry_request(
            session,
            input_identifier.value,
            input_identifier_cas,
            output_representations,
        )

        inv_identifier_map = inverse_map(OUTPUT_IDENTIFIER_MAP)
        return [
            CompoundIdentifier(
                identifier_type=inv_identifier_map[output_representation],
                value=result,
            )
            for output_representation, result in results.items()
        ]


async def common_chemistry_request(
    session: ClientSession,
    identifier: str,
    input_representation: str,
    output_representations: List[str],
) -> str:
    api_url = BASE_API + "/default/detail"
    params = {input_representation: identifier}

    logger.debug(f"Request: {api_url}")
    async with session.get(api_url, params=params) as resp:
        logger.debug(f"Response status: {resp.status}")
        if resp.status == 404:
            raise HTTPNotFound(text=f"Failed request for {identifier}")
        response = await resp.json()
    return {
        output_representation: response.get(output_representation)
        for output_representation in output_representations
    }
    # return response.get(output_representation)



from typing import List,Tuple,Dict
from .reaction import ReactionIdentifier, ReactionIdentifierType
from rdkit import Chem
from rdkit.Chem import PeriodicTable, GetPeriodicTable
from rdkit.Chem.rdMolDescriptors import CalcMolFormula
import ray

class CustomError(Exception):
    '''
    For error handling *Change/adapt as needed*
    '''
    pass


def initray(restart=True,num_cpus=16,log_to_driver=False):
    """
    Initializes a Ray CPU cluster for parallel execution of functions

    Args:
        restart (bool, optional): Restart existing cluster. Defaults to True.
        num_cpus (int, optional): Number of CPUs for parallel execution. Defaults to 16.
        log_to_driver (bool, optional): Controls whether results are logged to driver. Defaults to False.
    """
    if restart:
        ray.shutdown()
    ray.init(num_cpus=num_cpus,log_to_driver=log_to_driver)
    
def molfromsmiles(SMILES):
    """
    Converts a smiles string into a mol object for RDKit use. Returns a mol object. *Unsure how to integrate in compound.py*

    Args:
        SMILES (str): SMILES string to convert to mol object

    Returns:
        mol: RDKit mol object
    """
    mol=Chem.MolFromSmiles(SMILES)
    Chem.SanitizeMol(mol)
    mol.UpdatePropertyCache(strict=False)
    return mol
    
def atomtypes(mol)->Tuple[Dict,int]:
    """
    Generates an atom type dictionary with counts for a given mol file. Also
    returns overall change of mol. 

    Args:
        mol (RDKit mol): RDKit mol object to generate atom type dictionary for

    Returns:
        Tuple[Dict,int]: Dictionary with keys as atom element and value as count, overall charge
    """

    typedict={}
    mol2=Chem.AddHs(mol) #Assumes hydrogens are added
    charge=0
    for atom in mol2.GetAtoms():
        elem=PeriodicTable.GetElementSymbol(GetPeriodicTable(),atom.GetAtomicNum())
        charge+=atom.GetFormalCharge()
        if elem not in typedict.keys():
            typedict[elem]=1
        else:
            typedict[elem]+=1
    return typedict, charge



def getcompdict(ID=1,mol=None,smiles=None,formula=None)->Dict:
    """
    Wrapper for atomtypes. ID must be input whether Reaxys ID or random number. Default is 1. 
    Smiles and mol can also be specified. Can be converted to an object instance if intended.

    Args:
        ID (int, optional): ID of compound/mol. Defaults to 1.
        mol (RDKit mol, optional): RDKit mol. Defaults to None.
        smiles (str, optional): SMILES string. Defaults to None.
        formula (str, optional): Molecular formula. Defaults to None.

    Raises:
        CustomError: Raises error if smiles specified
        is not valid/cannot be processed

    Returns:
        Dict: Compound dictionary with ID as key and another dictionary with the
        following keys: 'atomdict' (Output of atomtypes, dictionary with atom
        elements and count), 'charge' (Overall charge of mol),'smiles'
        (smiles of mol),'formula' (chemical formula),'count' (instance number, always 1)
    """

    if smiles is not None:
        try:
            mol=molfromsmiles(smiles)
        except Exception as e:
            raise CustomError("Compound "+str(ID)+" smiles is invalid")
    if formula is None:
        formula=CalcMolFormula(mol)
    atomdata=atomtypes(mol)
    compddict={ID:{'atomdict':atomdata[0],'charge':atomdata[1],'smiles':smiles,'formula':formula,'count':1}}
    return compddict



def balance_reactions(
    reactions: List[ReactionIdentifier],
    ignore_hydrogens: bool = True,
    use_ray: bool = False,
    
) -> List[Tuple[ReactionIdentifier, ReactionIdentifier]]:
    if use_ray:

        @ray.remote
        def balance_reaction_ray(**kwargs):
            balance_reaction(**kwargs)

        for reaction in reactions:
            ray.remote(balance_reaction_ray)
    else:
        [balance_reaction(reaction) for reaction in reactions]


def balance_reaction(
    input_reaction_identifier: ReactionIdentifier,
) -> Tuple[ReactionIdentifier, ReactionIdentifier]:
    pass

    # do balancing
    output_reaction_identifier = None
    return input_reaction_identifier, output_reaction_identifier



from pura.services import Service
from pura.compound import CompoundIdentifier, CompoundIdentifierType
from aiohttp import ClientSession
from aiohttp.web_exceptions import HTTPNotFound
from typing import List, Union
from urllib.parse import quote
import logging


BASE_API = "https://opsin.ch.cam.ac.uk/opsin/"

IDENTIFIER_MAP = {
    CompoundIdentifierType.SMILES: "smiles",
    CompoundIdentifierType.INCHI: "stdinchi",
    CompoundIdentifierType.INCHI_KEY: "stdinchikey",
}

logger = logging.getLogger(__name__)


class Opsin(Service):
    async def resolve_compound(
        self,
        session: ClientSession,
        input_identifier: CompoundIdentifier,
        output_identifier_types: List[CompoundIdentifierType],
    ) -> List[CompoundIdentifier]:
        representations = [
            IDENTIFIER_MAP.get(output_identifier_type)
            for output_identifier_type in output_identifier_types
        ]
        if not any(representations):
            raise ValueError(
                f"{output_identifier_types} is not one of the valid identifier types for Opsin."
            )

        output_compound_identifiers = []
        for representation, output_identifier_type in zip(
            representations, output_identifier_types
        ):
            result = await opsin_request(
                session, input_identifier.value, representation
            )

            if result is not None:
                output_compound_identifiers += [
                    CompoundIdentifier(
                        identifier_type=output_identifier_type,
                        value=result,
                    )
                ]
        return output_compound_identifiers


async def opsin_request(
    session: ClientSession, name: str, output_representation: str
) -> str:
    api_url = BASE_API + quote(name)

    logger.debug(f"Request: {api_url}")
    async with session.get(api_url) as resp:
        logging.debug(f"Response status: {resp.status}")
        if resp.status == 404:
            raise HTTPNotFound(text=f"Failed request for {name}")
        response = await resp.json()
    return response.get(output_representation)



from pura.compound import (
    CompoundIdentifier,
    CompoundIdentifierType,
    Compound,
    standardize_identifier,
    unique_identifiers,
)
from pura.services import *
from tqdm import tqdm
from aiohttp import *
from aiohttp.web_exceptions import (
    HTTPClientError,
    HTTPServerError,
    HTTPServiceUnavailable,
)
import asyncio
import nest_asyncio

nest_asyncio.apply()
from typing import Optional, List, Union, Tuple, Dict, Callable
from itertools import combinations
from functools import reduce
import logging
import queue

logger = logging.getLogger(__name__)

aiohttp_errors = (
    HTTPServiceUnavailable,
    ClientConnectionError,
    TimeoutError,
    ClientConnectorCertificateError,
    ClientConnectorError,
    ClientConnectorSSLError,
    ClientError,
    ClientHttpProxyError,
    ClientOSError,
    ClientPayloadError,
    ClientProxyConnectionError,
    ClientResponseError,
    ClientSSLError,
    ContentTypeError,
    InvalidURL,
    ServerConnectionError,
    ServerDisconnectedError,
    ServerFingerprintMismatch,
    ServerTimeoutError,
    WSServerHandshakeError,
    asyncio.TimeoutError,
)


def base_check_agreement(
    identifiers_list: List[List[CompoundIdentifier]],
    agreement: int,
    service: Service,
) -> Tuple[List[List[CompoundIdentifier]], bool]:
    """
    Check if identifier lists from multiple services agree.

    Parameters
    ---------

    identifiers_list :  list of list of :class:`~pura.compound.CompoundIdentifier`
        List of lists of CompoudIdentifier objects. Each item in the first list is
        considered results from a  different service.
    agreement : int
        The number of services that must agree.
    service : Service
        The service used at the time of calling this method. Can be used
        by methods that override for further filtering.

    Returns
    -------

    A tuple of the following two objects:
    1. list of lists of CompoundIdentifier objects representing the unique identifiers
    2. boolean that is true if there is sufficient agreement between services

    Notes
    ------

    Algorithm:
    1. Find all combinatons of services that can satisfy agreement (combinations)
    2. Find the intersection of each combination
    3. If the intersection is greater than zero, then you have sufficient agreement.

    """
    if agreement <= 0:
        raise ValueError("Agreement must be greater than 0.")
    identifiers_list_new = []
    for identifiers in identifiers_list:
        if len(identifiers) > 0:
            identifiers_list_new.append(
                [identifier.value for identifier in identifiers]
            )
            identifier_type = identifiers[0].identifier_type
    identifiers_sets = [set(ident) for ident in identifiers_list_new]
    options = list(range(len(identifiers_list_new)))
    intersection = []
    for combo in combinations(options, agreement):
        intersection = reduce(
            set.intersection, [identifiers_sets[combo_i] for combo_i in combo]
        )
        if len(intersection) > 0:
            break
    if len(intersection) == 0:
        return identifiers_list, False
    else:
        return [
            [
                CompoundIdentifier(identifier_type=identifier_type, value=identifier)
                for identifier in intersection
            ]
        ], True


class CompoundResolver:
    """Resolve compound identifier types using external services such as PubChem.

    Parameters
    ----------

    services : list of :class:`~pura.service.Service`
        The services used for resolution.
    silent : bool, optional
        If True, logs errors but does not raise them. Default is False
    agreement_check : callable, optional
        A function that checks for agreement. See :class:`~pura.resolvers.base_check_agreement`
        for the API and the default agreement_check function.

    Examples
    --------

    The example below demonstrates how multiple identifiers of the same compound
    can be used to help increase the chances of resolution.  Each pairing of input identifier
    and service will be tried until sufficient agreement is reached or the pairings are exhausted.

    >>> services = [CAS(), CIR(), Opsin(), PubChem()]
    >>> compounds = [
    ...    Compound(
    ...            identifiers=[
    ...                CompoundIdentifier(
    ...                    identifier_type=CompoundIdentifierType.CAS_NUMBER,
    ...                    value="6737-11-7",
    ...                ),
    ...                CompoundIdentifier(
    ...                    identifier_type=CompoundIdentifierType.NAME,
    ...                    value="i-Butyric acid 1,2,3-propanetriyl ester",
    ...                ),
    ...            ]
    ...        )
    ... ]
    >>> resolver = CompoundResolver(services=services, silent=True)
    >>> compound_identifiers_list = resolver.resolve(
    ...        input_compounds=compounds,
    ...        output_identifier_type=CompoundIdentifierType.SMILES,
    ...        agreement=2,
    ...        batch_size=1,
    ... )
    >>> print(compound_identifiers_list)

    """

    def __init__(
        self,
        services: List[Service],
        silent: Optional[bool] = False,
        agreement_check: Optional[Callable] = None,
    ):
        self._services = services
        self.silent = silent
        self.agreement_check = (
            agreement_check if agreement_check is not None else base_check_agreement
        )

    def resolve(
        self,
        input_compounds: List[Compound],
        output_identifier_type: CompoundIdentifierType,
        backup_identifier_types: Optional[List[CompoundIdentifierType]] = None,
        agreement: Optional[int] = 1,
        batch_size: Optional[int] = None,
        n_retries: Optional[int] = 3,
    ) -> List[Tuple[Compound, Union[List[CompoundIdentifier], None]]]:
        """Resolve a list of compound identifiers to another identifier type(s).

        Parameters
        ---------

        input_compounds : list of :class:`~pura.compund.Compound`
            The list of Compounds that should be resolved.
        output_identifiers_types : list of :class:`~pura.compund.CompoundIdentifier`
            The compound identifier type that should be outputted.
        backup_identifier_types : list of :class:`~pura.compound.CompoundIdentifierType`
            A list of identifier types that can be looked up and then used to resolve
            to the desired output identifier type. Default is None.
        agreement : int, optional
            The number of services that must give the same resolved
            CompoundIdentifier for the resolution to be considered correct.
            If set to zero, then all services will be tried and the unique
            results returned. Default is 1.
        batch_size : int, optional
            The batch size sets the number of requests to send simultaneously.
            Defaults to 10 or the length input_idententifier, whichever is smaller.
        n_retries : int, optional
            The number of times a request should be retried if there is a problem
            in contacting a service (e.g., an internet outage). Defaults to 3.


        Notes
        -----

        The retries option uses an exponential backoff, so a sleep of 2^n seconds will occur before
        the next retry, where n is the number of retries thus far.

        Returns
        -------

        A list of tuples where the first element of each tuple is the input compound and the second
        element is a list of compound identifier(s).

        """
        # Make sure output identifier type is different than backup identifier types
        if (
            backup_identifier_types is not None
            and output_identifier_type in backup_identifier_types
        ):
            raise ValueError(
                "Output identifier type cannot be in backup identifier types."
            )
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError as e:
            if str(e).startswith("There is no current event loop in thread"):
                loop = asyncio.new_event_loop()
                asyncio.set_event_loop(loop)
            else:
                raise
        return loop.run_until_complete(
            self._resolve(
                input_compounds=input_compounds,
                output_identifier_type=output_identifier_type,
                backup_identifier_types=backup_identifier_types,
                agreement=agreement,
                batch_size=batch_size,
                n_retries=n_retries,
            )
        )

    async def _resolve(
        self,
        input_compounds: List[Compound],
        output_identifier_type: CompoundIdentifierType,
        backup_identifier_types: Optional[List[CompoundIdentifierType]] = None,
        agreement: Optional[int] = 1,
        batch_size: Optional[int] = None,
        n_retries: Optional[int] = 3,
    ) -> List[Tuple[Compound, Union[List[CompoundIdentifier], None]]]:
        """This is the async function with the same API as resolve"""

        n_identifiers = len(input_compounds)
        if batch_size is None:
            batch_size = 10 if n_identifiers >= 10 else n_identifiers
        n_batches = n_identifiers // batch_size
        n_batches += 0 if n_identifiers % batch_size == 0 else 1
        resolved_identifiers = []
        backup_identifier_types = (
            backup_identifier_types if backup_identifier_types is not None else []
        )
        # Iterate through batches
        for batch in tqdm(range(n_batches), position=0, desc="Batch"):
            # Get subset of data
            start = batch * batch_size
            batch_identifiers = input_compounds[start : start + batch_size]

            # Start aiohttp session
            async with ClientSession() as session:
                # Create series of tasks to run in parallel
                tasks = [
                    self._resolve_one_compound(
                        session,
                        compound_identifier,
                        output_identifier_type,
                        backup_identifier_types,
                        agreement,
                        n_retries=n_retries,
                    )
                    for compound_identifier in batch_identifiers
                ]
                batch_bar = tqdm(
                    asyncio.as_completed(tasks),
                    total=len(tasks),
                    desc=f"Batch {batch} Progress",
                    position=1,
                    leave=True,
                )
                resolved_identifiers.extend([await f for f in batch_bar])
                batch_bar.clear()

        return resolved_identifiers

    async def _resolve_one_compound(
        self,
        session: ClientSession,
        input_compound: Compound,
        output_identifier_type: CompoundIdentifierType,
        backup_identifier_types: List[CompoundIdentifierType],
        agreement: int,
        n_retries: Optional[int],
    ) -> Tuple[Compound, Union[List[CompoundIdentifier], None]]:
        """Resolve one compound"""
        resolved_identifiers_list = []
        agreement_satisfied = False
        # Create input identifier queue
        input_identifiers_queue = queue.Queue()
        input_identifiers_list = [
            (identifier, None) for identifier in input_compound.identifiers
        ]
        # for identifier in input_compound.identifiers:
        #     input_identifiers_queue.put()

        # Main loop
        while len(input_identifiers_list) > 0 and not agreement_satisfied:
            input_identifier, no_go_service = input_identifiers_list[0]
            input_identifiers_list = input_identifiers_list[1:]
            for service in self._services:
                if service == no_go_service:
                    continue
                for j in range(n_retries):
                    try:
                        output_identifier_types = [
                            output_identifier_type
                        ] + backup_identifier_types
                        if input_identifier.identifier_type in output_identifier_types:
                            output_identifier_types.remove(
                                input_identifier.identifier_type
                            )
                        resolved_identifiers = await service.resolve_compound(
                            session,
                            input_identifier=input_identifier,
                            output_identifier_types=output_identifier_types,
                        )
                        logger.debug(
                            f"{service} | {input_identifier.value}->{resolved_identifiers}"
                        )

                        final_resolved_identifiers = []
                        for identifier in resolved_identifiers:
                            # Add backup identifiers to queue
                            if identifier.identifier_type in backup_identifier_types:
                                if (
                                    (identifier, service) not in input_identifiers_list
                                    # But not second order backup identifier checks
                                    and input_identifier.identifier_type
                                    not in backup_identifier_types
                                ):
                                    input_identifiers_list.append((identifier, service))
                                continue

                            # Standardize identifiers (e.g., SMILES canonicalization)
                            if identifier is not None:
                                standardize_identifier(identifier)

                            final_resolved_identifiers.append(identifier)

                        resolved_identifiers_list.append(final_resolved_identifiers)
                        break
                    except aiohttp_errors as e:  # type: ignore
                        # If server is busy, use exponential backoff
                        logger.debug(f"Sleeping for {2**j} ({e})")
                        await asyncio.sleep(2**j)
                    except (HTTPClientError, HTTPServerError) as e:
                        # Log/raise on all other HTTP errors
                        if self.silent:
                            logger.error(msg=f"{service}, {input_identifier}: {e}")
                            break
                        else:
                            raise e
                    except ValueError as e:
                        if self.silent:
                            logger.error(e)
                            break
                        else:
                            raise e

                # Check agreement between services
                if len(resolved_identifiers_list) > 0 and agreement > 1:
                    (
                        resolved_identifiers_list,
                        agreement_satisfied,
                    ) = self.agreement_check(
                        resolved_identifiers_list,
                        agreement,
                        service,
                    )
                elif (
                    agreement == 1
                    and len(resolved_identifiers_list) > 0
                    and len(resolved_identifiers_list[0]) > 0
                ):
                    agreement_satisfied = True

                if agreement_satisfied:
                    break

        # If agreement is 0 or 1, then we just want to dedup and return
        if agreement == 0 or agreement == 1:
            resolved_identifiers_list = [
                unique_identifiers(resolved_identifiers)
                for resolved_identifiers in resolved_identifiers_list
            ]
            agreement_satisfied = True

        if not agreement_satisfied:
            error_txt = f"Not sufficient agreement for {input_compound} (outputs: {resolved_identifiers_list})"
            if self.silent:
                logger.error(error_txt)
                resolved_identifiers_list = []
            else:
                raise ResolverError(error_txt)
        return input_compound, flatten_list(resolved_identifiers_list)


def flatten_list(l: List):
    """Flatten a multidimensional list to one dimension"""
    lnew = []
    for li in l:
        if type(li) != list:
            lnew.append(li)
        else:
            lnew.extend(flatten_list(li))
    return lnew


class ResolverError(Exception):
    """General error problems with resolving names."""

    pass


def resolve_identifiers(
    names: List[str],
    output_identifier_type: CompoundIdentifierType,
    input_identifer_type: CompoundIdentifierType = CompoundIdentifierType.NAME,
    backup_identifier_types: Optional[List[CompoundIdentifierType]] = None,
    agreement: int = 1,
    batch_size: int = 100,
    services: Optional[List[Service]] = None,
    silent: Optional[bool] = False,
) -> List[Tuple[str, Union[List[str], None]]]:
    """Resolve a list of names (or any other identifier) to an identifier type.

    Parameters
    -----------

    names : list of str
        The list of compound names that should be resolved
    output_identifier_type : :class:`~pura.compund.CompoundIdentifierType`
        The list of compound identifier types to resolve to.
    input_identifier_type : :class:`~pura.compund.CompoundIdentifierType`, optional
        The input identifier type, Defaults to name
    backup_identifier_types : list of :class:`~pura.compound.CompoundIdentifierType`
        A list of identifier types that can be looked up and then used to resolve
        to the desired output identifier type. Default is None.
    agreement : int, optional
        The number of services that must give the same resolved
        `CompoundIdentifier` for the resolution to be considered correct.
        If set to zero, then all services will be tried and the unique
        results returned. Default is 1.
    batch_size : int, optional
        The batch size sets the number of requests to send simultaneously.
        Defaults to 100 or the length input_idententifier, whichever is smaller.
    services : list of :class:`~pura.service.Service`, optional
        Services used to do resolution. Defaults to PubChem and CIR.
    silent : bool, optional
        If True, logs errors but does not raise them. Default is False

    Returns
    -------
    List of tuples where the first element in each tuple is the input identifier
    and the second element is a list of resolved identifiers.


    Example
    -------

    >>> from pura.services import  Pubchem, CIR
    >>> smiles = resolve_identifiers(
    ...     ["aspirin", "ibuprofen", "toluene"],
    ...     output_identifier_type=CompoundIdentifierType.SMILES,
    ...     services=[Pubchem(), CIR()],
    ...     agreement=2,
    ... )

    Notes
    -----

    This is a convenience function for quickly resolving a list of strings without having
    to create Compound objects.

    Due to the asynchronous nature of the API calls, the function will not
    return the results in the same order as the input list.

    """
    if services is None:
        services = [PubChem(autocomplete=True), CIR()]
    compounds = [
        Compound(
            identifiers=[
                CompoundIdentifier(identifier_type=input_identifer_type, value=name)
            ]
        )
        for name in names
    ]
    resolver = CompoundResolver(services=services, silent=silent)
    results = resolver.resolve(
        input_compounds=compounds,
        output_identifier_type=output_identifier_type,
        backup_identifier_types=backup_identifier_types,
        agreement=agreement,
        batch_size=batch_size,
    )
    return [
        (
            input_compound.identifiers[0].value,
            [identifier.value for identifier in resolved_identifiers],
        )
        for input_compound, resolved_identifiers in results
    ]



from .compound import Compound

# from .units import Time, Temperature
from typing import List, Tuple, Optional, Dict
import pandas as pd
from enum import Enum
from pydantic import BaseModel


class ReactionIdentifierType(BaseModel):
    """Generic identifiers for a reaction"""

    UNSPECIFIED = 0
    CUSTOM = 1
    REACTION_SMILES = 2
    REACTION_CXSMILES = 6
    # Extended SMILES.
    RDFILE = 3
    # Reaction data file.
    RINCHI = 4
    # Reaction InChI.
    NAME = 5
    # Named reaction or reaction category.


class ReactionRole(BaseModel):
    UNSPECIFIED = 0
    # A reactant is any compound that contributes atoms to a desired or
    # observed product.
    REACTANT = 1
    REAGENT = 2
    SOLVENT = 3
    CATALYST = 4
    # The workup role is used when defining quenches, buffer additives for
    # liquid-liquid separations, etc.
    WORKUP = 5
    # Internal standards can be included as part of a reaction input (when
    # added prior to the start of the reaction) or as part of a workup
    # step of addition.
    INTERNAL_STANDARD = 6
    AUTHENTIC_STANDARD = 7
    # A product can be any species produced by the reaction, whether desired
    # or undesired.
    PRODUCT = 8


class ReactionIdentifier(BaseModel):
    identifier_type: ReactionIdentifierType
    value: str
    details: Optional[str] = None


class ReactionInput(BaseModel):
    # This is different than ORD. We don't include ReactionRole inside
    # the Compound object because compounds can exist independent of reactions
    components: Dict[ReactionRole, Compound]
    addition_order: Optional[int] = None
    # addition_time: Optional[Time] = None
    # addition_duration: Optional[Time] = None
    flow_rate: Optional[float] = None
    # addition_temperature: Optional[Temperature] = None


class ReactionConditions(BaseModel):
    pass


class Reaction(BaseModel):
    """

    Reaction identifiers are the whole reaction and should be convertible
    to inputs + outcomes.
    """

    identifiers: List[ReactionIdentifier]

    # List of pure substances or mixtures that were added to the reaction vessel.
    # This is a map instead of a repeated field to simplify reaction templating
    # through the use of keys. String keys are simple descriptions and are
    # present only for convenience.
    inputs: List[ReactionInput]
    conditions: ReactionConditions

    def to_row(self):
        return


def to_frame(reactions: List[Reaction], columns: List[str]) -> pd.DataFrame:
    reaction_dicts = [reaction.to_row() for reaction in reactions]
    return pd.DataFrame(reaction_dicts)



"""
Python interface for the Chemical Identifier Resolver (CIR) by the CADD Group at the NCI/NIH.
Based heavily on CIRpy by Matt Swain
https://github.com/mcs07/CIRpy

"""
from pura.services import Service
from pura.compound import CompoundIdentifier, CompoundIdentifierType
from pura.utils import inverse_map
from aiohttp import ClientSession
from typing import List, Union
import logging

try:
    from urllib.error import HTTPError
    from urllib.parse import quote, urlencode
    from urllib.request import urlopen
except ImportError:
    from urllib import urlencode
    from urllib2 import quote, urlopen, HTTPError

try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.cElementTree as etree
    except ImportError:
        import xml.etree.ElementTree as etree


logger = logging.getLogger(__name__)

API_BASE = "https://cactus.nci.nih.gov/chemical/structure"


FILE_FORMATS = {
    "alc",
    "cdxml",
    "cerius",
    "charmm",
    "cif",
    "cml",
    "ctx",
    "gjf",
    "gromacs",
    "hyperchem",
    "jme",
    "maestro",
    "mol",
    "mol2",
    "mrv",
    "pdb",
    "sdf3000",
    "sln",
    "xyz",
}
IDENTIFIER_MAP = {
    CompoundIdentifierType.SMILES: "smiles",
    CompoundIdentifierType.MOLBLOCK: "mol",
    CompoundIdentifierType.INCHI: "stdinchi",
    CompoundIdentifierType.IUPAC_NAME: "iupac_name",
    CompoundIdentifierType.CAS_NUMBER: "cas",
    CompoundIdentifierType.INCHI_KEY: "stdinchikey",
    CompoundIdentifierType.XYZ: "xyz",
}


class CIR(Service):
    async def resolve_compound(
        self,
        session: ClientSession,
        input_identifier: CompoundIdentifier,
        output_identifier_types: List[CompoundIdentifierType],
    ) -> List[Union[CompoundIdentifierType, None]]:
        representations = [
            IDENTIFIER_MAP.get(output_identifier_type)
            for output_identifier_type in output_identifier_types
        ]
        if not any(representations):
            raise ValueError(
                f"{output_identifier_types} is not one of the valid identifier types for the chemical identifier resolver."
            )

        output_compound_identifiers = []
        for representation, output_identifier_type in zip(
            representations, output_identifier_types
        ):
            if representation is None:
                continue
            values = await resolve(
                session, input=input_identifier.value, representation=representation
            )
            if isinstance(values, str):
                values = [values]

            if len(values) > 0:
                output_compound_identifiers += [
                    CompoundIdentifier(
                        identifier_type=output_identifier_type, value=value
                    )
                    for value in values
                ]

        return output_compound_identifiers


async def resolve(
    session: ClientSession,
    input: str,
    representation: str,
    resolvers: List[str] = None,
    get3d: bool = False,
    **kwargs,
):
    """Resolve input to the specified output representation.

    :param aiohttp.ClientSession() session: Aiohttp session
    :param string input: Chemical identifier to resolve
    :param string representation: Desired output representation
    :param list(string) resolvers: (Optional) Ordered list of resolvers to use
    :param bool get3d: (Optional) Whether to return 3D coordinates (where applicable)
    :returns: Output representation or None
    :rtype: string or None
    :raises HTTPError: if CIR returns an error code
    :raises ParseError: if CIR response is uninterpretable
    """
    # Take first result from XML query
    results = await query(
        session, input, representation, resolvers, False, get3d, **kwargs
    )
    result = results[0].value if results else None
    if result is None:
        return []
    elif not isinstance(result, list):
        return [result]
    else:
        return result


class Result(object):
    """A single result returned by CIR."""

    def __init__(self, input, notation, input_format, resolver, representation, value):
        """

        :param string input: Originally supplied input identifier that produced this result
        :param string notation: Identifier matched by the resolver or tautomer ID
        :param string input_format: Format of the input as interpreted by the resolver
        :param string resolver: Resolver used to produce this result
        :param string representation: Requested output representation
        :param value: Actual result value
        :type value: string or list(string)
        """
        self.input = input
        self.representation = representation
        self.resolver = resolver
        self.input_format = input_format
        self.notation = notation
        self.value = value

    def __repr__(self):
        return (
            "Result(input=%r, representation=%r, resolver=%r, input_format=%r, notation=%r, value=%r)"
            % (
                self.input,
                self.representation,
                self.resolver,
                self.input_format,
                self.notation,
                self.value,
            )
        )

    def __str__(self):
        return self.value

    def __eq__(self, other):
        return isinstance(other, type(self)) and self.__dict__ == other.__dict__

    def __getitem__(self, prop):
        """Allow dict-style access to attributes to ease transition from when results were dicts."""
        if prop in self.__dict__:
            return getattr(self, prop)
        raise KeyError(prop)

    def __setitem__(self, prop, val):
        """Allow dict-style setting of attributes to ease transition from when results were dicts."""
        setattr(self, prop, val)

    def __contains__(self, prop):
        """Allow dict-style checking of attributes to ease transition from when results were dicts."""
        return prop in self.__dict__

    def to_dict(self):
        """Return a dictionary containing Result data."""
        return self.__dict__


async def query(
    session,
    input,
    representation,
    resolvers=None,
    get3d=False,
    tautomers=False,
    **kwargs,
):
    """Get all results for resolving input to the specified output representation.

    :param aiohttp.ClientSession() session: Aiohttp session
    :param string input: Chemical identifier to resolve
    :param string representation: Desired output representation
    :param list(string) resolvers: (Optional) Ordered list of resolvers to use
    :param bool get3d: (Optional) Whether to return 3D coordinates (where applicable)
    :param bool tautomers: (Optional) Whether to return all tautomers
    :returns: List of resolved results
    :rtype: list(Result)
    :raises HTTPError: if CIR returns an error code
    :raises ParseError: if CIR response is uninterpretable
    """
    tree = await request(
        session, input, representation, resolvers, get3d, tautomers, **kwargs
    )

    results = []
    if tree is None:
        return results
    for data in tree.findall(".//data"):
        value = [item.text for item in data.findall("item")]
        result = Result(
            input=tree.attrib["string"],
            representation=tree.attrib["representation"],
            resolver=data.attrib["resolver"],
            input_format=data.attrib["string_class"],
            notation=data.attrib["notation"],
            value=value[0] if len(value) == 1 else value,
        )
        results.append(result)
    logger.debug("Received %s query results", len(results))
    return results


async def request(
    session: ClientSession,
    input: str,
    representation: str,
    resolvers: List[str] = None,
    get3d: bool = False,
    tautomers: bool = False,
    **kwargs,
):
    """Make a request to CIR and return the XML response.

    :param aiohttp.ClientSession() session: Aiohttp session
    :param string input: Chemical identifier to resolve
    :param string representation: Desired output representation
    :param list(string) resolvers: (Optional) Ordered list of resolvers to use
    :param bool get3d: (Optional) Whether to return 3D coordinates (where applicable)
    :param bool tautomers: (Optional) Whether to return all tautomers
    :returns: XML response from CIR
    :rtype: Element
    :raises HTTPError: if CIR returns an error code
    :raises ParseError: if CIR response is uninterpretable
    """
    url = construct_api_url(
        input, representation, resolvers, get3d, tautomers, **kwargs
    )
    logger.debug("Making request: %s", url)
    async with session.get(url) as resp:
        response = await resp.text()
    try:
        feed = etree.fromstring(response.encode("ascii"))
        tree = etree.ElementTree(feed)
        return tree.getroot()
    except (UnicodeDecodeError, UnicodeEncodeError) as e:
        logger.error(e)
        return


def construct_api_url(
    input,
    representation,
    resolvers=None,
    get3d=False,
    tautomers=False,
    xml=True,
    **kwargs,
):
    """Return the URL for the desired API endpoint.

    :param string input: Chemical identifier to resolve
    :param string representation: Desired output representation
    :param list(str) resolvers: (Optional) Ordered list of resolvers to use
    :param bool get3d: (Optional) Whether to return 3D coordinates (where applicable)
    :param bool tautomers: (Optional) Whether to return all tautomers
    :param bool xml: (Optional) Whether to return full XML response
    :returns: CIR API URL
    :rtype: str
    """
    # File formats require representation=file and the format in the querystring
    if representation in FILE_FORMATS:
        kwargs["format"] = representation
        representation = "file"
    # Prepend input with 'tautomers:' to return all tautomers
    if tautomers:
        input = "tautomers:%s" % input
    url = "%s/%s/%s" % (API_BASE, quote(input), representation)
    if xml:
        url += "/xml"
    if resolvers:
        kwargs["resolver"] = ",".join(resolvers)
    if get3d:
        kwargs["get3d"] = True
    if kwargs:
        url += "?%s" % urlencode(kwargs)
    return url



# -*- coding: utf-8 -*-
"""
PubChemPy

Python interface for the PubChem PUG REST service.
https://github.com/mcs07/PubChemPy
"""
from queue import Queue
from urllib.parse import quote, urlencode
from pura.services import Service
from pura.compound import CompoundIdentifier, CompoundIdentifierType
from pura.utils import inverse_map
from aiohttp import ClientSession
from aiohttp.web_exceptions import (
    HTTPBadRequest,
    HTTPNotFound,
    HTTPServiceUnavailable,
    HTTPMethodNotAllowed,
    HTTPGatewayTimeout,
    HTTPNotImplemented,
    HTTPInternalServerError,
)
from typing import List, Optional, Union
import logging


logger = logging.getLogger(__name__)


text_types = str, bytes


API_BASE = "https://pubchem.ncbi.nlm.nih.gov/rest/pug"
AUTOCOMPLETE_BASE = "https://pubchem.ncbi.nlm.nih.gov/rest/autocomplete"

INPUT_IDENTIFIER_MAP = {
    CompoundIdentifierType.SMILES: "CanonicalSMILES",
    # CompoundIdentifierType.MOLBLOCK: "mol",
    CompoundIdentifierType.INCHI: "InChI",
    CompoundIdentifierType.IUPAC_NAME: "IUPACName",
    CompoundIdentifierType.INCHI_KEY: "InChIKey",
    # CompoundIdentifierType.XYZ: "xyz",
    CompoundIdentifierType.NAME: "name",
}

OUTPUT_IDENTIFIER_MAP = {
    CompoundIdentifierType.SMILES: "CanonicalSMILES",
    CompoundIdentifierType.INCHI: "InChI",
    CompoundIdentifierType.IUPAC_NAME: "IUPACName",
    CompoundIdentifierType.INCHI_KEY: "InChIKey",
}

# Allows properties to optionally be specified as underscore_separated, consistent with Compound attributes
PROPERTY_MAP = {
    "molecular_formula": "MolecularFormula",
    "molecular_weight": "MolecularWeight",
    "canonical_smiles": "CanonicalSMILES",
    "isomeric_smiles": "IsomericSMILES",
    "inchi": "InChI",
    "inchikey": "InChIKey",
    "iupac_name": "IUPACName",
    "xlogp": "XLogP",
    "exact_mass": "ExactMass",
    "monoisotopic_mass": "MonoisotopicMass",
    "tpsa": "TPSA",
    "complexity": "Complexity",
    "charge": "Charge",
    "h_bond_donor_count": "HBondDonorCount",
    "h_bond_acceptor_count": "HBondAcceptorCount",
    "rotatable_bond_count": "RotatableBondCount",
    "heavy_atom_count": "HeavyAtomCount",
    "isotope_atom_count": "IsotopeAtomCount",
    "atom_stereo_count": "AtomStereoCount",
    "defined_atom_stereo_count": "DefinedAtomStereoCount",
    "undefined_atom_stereo_count": "UndefinedAtomStereoCount",
    "bond_stereo_count": "BondStereoCount",
    "defined_bond_stereo_count": "DefinedBondStereoCount",
    "undefined_bond_stereo_count": "UndefinedBondStereoCount",
    "covalent_unit_count": "CovalentUnitCount",
    "volume_3d": "Volume3D",
    "conformer_rmsd_3d": "ConformerModelRMSD3D",
    "conformer_model_rmsd_3d": "ConformerModelRMSD3D",
    "x_steric_quadrupole_3d": "XStericQuadrupole3D",
    "y_steric_quadrupole_3d": "YStericQuadrupole3D",
    "z_steric_quadrupole_3d": "ZStericQuadrupole3D",
    "feature_count_3d": "FeatureCount3D",
    "feature_acceptor_count_3d": "FeatureAcceptorCount3D",
    "feature_donor_count_3d": "FeatureDonorCount3D",
    "feature_anion_count_3d": "FeatureAnionCount3D",
    "feature_cation_count_3d": "FeatureCationCount3D",
    "feature_ring_count_3d": "FeatureRingCount3D",
    "feature_hydrophobe_count_3d": "FeatureHydrophobeCount3D",
    "effective_rotor_count_3d": "EffectiveRotorCount3D",
    "conformer_count_3d": "ConformerCount3D",
}


class PubChem(Service):
    """PubhChem is a service offered by the NIH in the United States.

    Paramaters
    ----------
    autocomplete: bool
        If the input identifier cannot be found, you can try autocomplete to have PubChem find alternative names.
        Default is False.

    Notes
    -----
    Pubchem can throttle with lots of requests: https://pubchemdocs.ncbi.nlm.nih.gov/dynamic-request-throttling

    """

    def __init__(self, autocomplete: bool = False, autocomplete_limit: int = 1) -> None:
        self.autocomplete = autocomplete
        self.autocomplete_limit = autocomplete_limit
        super().__init__()

    async def resolve_compound(
        self,
        session: ClientSession,
        input_identifier: CompoundIdentifier,
        output_identifier_types: List[CompoundIdentifierType],
    ) -> List[Union[CompoundIdentifierType, None]]:
        namespace = INPUT_IDENTIFIER_MAP.get(input_identifier.identifier_type)
        if namespace is None:
            raise ValueError(
                f"{input_identifier.identifier_type} is not one of the valid identifier types for Pubchem."
            )
        representations = [
            OUTPUT_IDENTIFIER_MAP.get(output_identifier_type)
            for output_identifier_type in output_identifier_types
        ]
        representations = list(set(representations))
        if None in representations:
            representations.remove(None)
        if not any(representations):
            raise ValueError(
                f"{output_identifier_types} contains invalid identifier types for PbuChme."
            )

        # Search
        input_identifiers_queue = Queue()
        input_identifiers_queue.put(input_identifier.value)
        autocomplete_tried = False
        while not input_identifiers_queue.empty():
            input_value = input_identifiers_queue.get()
            results = await get_properties(
                session,
                properties=representations,
                identifier=input_value,
                namespace=namespace,
                searchtype=None,
            )
            output_identifiers = []
            for representation in representations:
                for result in results:
                    if result and result.get(representation):
                        output_identifiers += [
                            CompoundIdentifier(
                                identifier_type=inverse_map(OUTPUT_IDENTIFIER_MAP)[
                                    representation
                                ],
                                value=result[representation],
                            )
                        ]

            # Autocomplete if search fails
            if (
                len(output_identifiers) == 0
                and self.autocomplete
                and not autocomplete_tried
            ):
                names = await autocomplete(
                    session, input_value, limit=self.autocomplete_limit
                )
                for name in names:
                    input_identifiers_queue.put(name)

        return output_identifiers


async def get_properties(
    session: ClientSession,
    properties,
    identifier,
    namespace="cid",
    searchtype=None,
    **kwargs,
):
    """Retrieve the specified properties from PubChem.

    :param identifier: The compound, substance or assay identifier to use as a search query.
    :param namespace: (optional) The identifier type.
    :param searchtype: (optional) The advanced search type, one of substructure, superstructure or similarity.
    :param as_dataframe: (optional) Automatically extract the properties into a pandas :class:`~pandas.DataFrame`.
    """
    if isinstance(properties, text_types):
        properties = properties.split(",")
    properties = ",".join([PROPERTY_MAP.get(p, p) for p in properties])
    properties = "property/%s" % properties

    try:
        results = await request(
            session=session,
            api_base=API_BASE,
            identifier=identifier,
            namespace=namespace,
            domain="compound",
            operation=properties,
            output="JSON",
            searchtype=searchtype,
            **kwargs,
        )
    except HTTPNotFound:
        return []

    if results is not None:
        logger.debug(results)
        return results["PropertyTable"]["Properties"]
    else:
        return []


async def autocomplete(
    session: ClientSession,
    identifier: str,
    lookup_dictionary: Optional[str] = "compound",
    **kwargs,
):
    # Call the autocomplete url
    try:
        results = await request(
            session=session,
            api_base=AUTOCOMPLETE_BASE,
            identifier=identifier,
            namespace=None,
            domain=lookup_dictionary,
            operation=identifier,
            output="JSON",
            searchtype=None,
            **kwargs,
        )
    except HTTPNotFound:
        return []

    if results.get("dictionary_terms"):
        logger.debug(results)
        return results["dictionary_terms"]["compound"]
    else:
        return []


async def request(
    session: ClientSession,
    api_base: str,
    identifier,
    namespace="cid",
    domain="compound",
    operation=None,
    output="JSON",
    searchtype=None,
    **kwargs,
):
    """
    Construct API request from parameters and return the response.

    Full specification at http://pubchem.ncbi.nlm.nih.gov/pug_rest/PUG_REST.html
    """
    if not identifier:
        raise ValueError("identifier/cid cannot be None")
    # If identifier is a list, join with commas into string
    if isinstance(identifier, int):
        identifier = str(identifier)
    if not isinstance(identifier, text_types):
        identifier = ",".join(str(x) for x in identifier)
    # Filter None values from kwargs
    kwargs = dict((k, v) for k, v in kwargs.items() if v is not None)
    # Build API URL
    urlid, postdata = None, {}
    if namespace == "sourceid":
        identifier = identifier.replace("/", ".")
    if (
        namespace in ["listkey", "formula", "sourceid"]
        or searchtype == "xref"
        or (searchtype and namespace == "cid")
        or domain == "sources"
    ):
        urlid = quote(identifier.encode("utf8"))
    else:
        # postdata = urlencode([(namespace, identifier)]).encode("utf8")
        postdata = {namespace: identifier}
    comps = filter(
        None, [api_base, domain, searchtype, namespace, urlid, operation, output]
    )
    apiurl = "/".join(comps)
    if kwargs:
        apiurl += "?%s" % urlencode(kwargs)

    # Make request
    logger.debug("Request URL: %s", apiurl)
    logger.debug("Request data: %s", postdata)
    async with session.post(apiurl, data=postdata) as resp:
        response = await resp.json()
    if response.get("Fault"):
        code = response["Fault"]["Code"]
        if code == "PUGREST.BadRequest":
            raise HTTPBadRequest(reason=code)
        elif code == "PUGREST.NotFound":
            raise HTTPNotFound(reason=code)
        elif code == "PUGREST.NotAllowed":
            raise HTTPMethodNotAllowed(reason=code)
        elif code == "PUGREST.Timeout":
            raise HTTPGatewayTimeout(reason=code)
        elif code == "PUGREST.ServerBusy":
            raise HTTPServiceUnavailable(reason=code)
        elif code == "PUGREST.Unimplemented":
            raise HTTPNotImplemented(reason=code)
        elif code == "PUGREST.ServerError" or code == "PUGREST.Unknown":
            raise HTTPInternalServerError()
    return response



from pura.compound import CompoundIdentifier, CompoundIdentifierType
from aiohttp import ClientSession
from abc import ABC, abstractmethod
from typing import List, Union


class Service(ABC):
    def __init__(self) -> None:
        pass

    @abstractmethod
    async def resolve_compound(
        self,
        session: ClientSession,
        input_identifier: CompoundIdentifier,
        output_identifier_types: List[CompoundIdentifierType],
    ) -> List[Union[CompoundIdentifier, None]]:
        pass



"""
Deep learning model for name and smiles resolution.

Based on code from Kohulan Rajan
https://github.com/Kohulan/Smiles-TO-iUpac-Translator
Paper: https://jcheminf.biomedcentral.com/articles/10.1186/s13321-021-00512-4

"""

from pura.services import Service
from pura.compound import CompoundIdentifier, CompoundIdentifierType
from typing import List, Union
from aiohttp import ClientSession
from rdkit import Chem

try:
    import tensorflow as tf
    import tensorflow.keras as keras
    import pystow

    STOUT_INSTALLED = True
except ImportError:
    STOUT_INSTALLED = False

import os
import pickle
import re
import logging
import re
import unicodedata
import numpy as np
import subprocess

logger = logging.getLogger(__name__)


class STOUT(Service):
    """
    STOUT is a deep learning model for name and smiles resolution.

    """

    def __init__(self) -> None:
        if not STOUT_INSTALLED:
            raise ImportError(
                "STOUT extra dependencies are not installed. Use poetry install -E stout"
            )
        os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

        # Print tensorflow version
        logger.debug("Tensorflow version: " + tf.__version__)

        # Always select a GPU if available
        os.environ["CUDA_VISIBLE_DEVICES"] = "0"

        # Scale memory growth as needed
        gpus = tf.config.experimental.list_physical_devices("GPU")
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)

        # Set path
        self.default_path = pystow.join("STOUT-V2", "models")

        # model download location
        model_url = "https://storage.googleapis.com/decimer_weights/models.zip"
        model_path = str(self.default_path) + "/translator_forward/"

        # download models to a default location
        if not os.path.exists(model_path):
            download_trained_weights(model_url, self.default_path)

        # Load the packed model forward
        logger.info("Loading STOUT models")
        self.reloaded_forward = tf.saved_model.load(
            self.default_path.as_posix() + "/translator_forward"
        )

        # Load the packed model forward
        self.reloaded_reverse = tf.saved_model.load(
            self.default_path.as_posix() + "/translator_reverse"
        )

    async def resolve_compound(
        self,
        session: ClientSession,
        input_identifier: CompoundIdentifier,
        output_identifier_types: List[CompoundIdentifierType],
    ) -> List[Union[CompoundIdentifier, None]]:
        if not (
            CompoundIdentifierType.IUPAC_NAME in output_identifier_types
            or CompoundIdentifierType.SMILES in output_identifier_types
        ):
            raise ValueError("STOUT can only resolve to IUPAC_NAME or SMILES")
        output_identifier_type = output_identifier_types[0]
        if (
            input_identifier.identifier_type == CompoundIdentifierType.SMILES
            and output_identifier_type == CompoundIdentifierType.IUPAC_NAME
        ):
            name = self.smiles_to_iupac(input_identifier.value)
            return [
                CompoundIdentifier(
                    identifier_type=CompoundIdentifierType.IUPAC_NAME, value=name
                )
            ]
        elif (
            input_identifier.identifier_type
            in [
                CompoundIdentifierType.IUPAC_NAME,
                CompoundIdentifierType.NAME,
            ]
            and output_identifier_type == CompoundIdentifierType.SMILES
        ):
            smiles = self.iupace_to_smiles(input_identifier.value)
            return [
                CompoundIdentifier(
                    identifier_type=CompoundIdentifierType.SMILES, value=smiles
                )
            ]
        else:
            raise ValueError(
                "Input identifier must be iupac/name or smiles and output must be iupac or smiles for STOUT"
            )

    def smiles_to_iupac(self, smiles: str) -> str:
        """Takes user input splits them into words and generates tokens.
        Tokens are then passed to the model and the model predicted tokens are retrieved.
        The predicted tokens gets detokenized and the final result is returned in a string format.
        Args:
            smiles (str): user input SMILES in string format.
        Returns:
            result (str): The predicted IUPAC names in string format.
        """

        # Load important pickle files which consists the tokenizers and the maxlength setting
        inp_lang = pickle.load(
            open(self.default_path.as_posix() + "/assets/tokenizer_input.pkl", "rb")
        )
        targ_lang = pickle.load(
            open(self.default_path.as_posix() + "/assets/tokenizer_target.pkl", "rb")
        )
        inp_max_length = pickle.load(
            open(self.default_path.as_posix() + "/assets/max_length_inp.pkl", "rb")
        )
        if len(smiles) == 0:
            return ""
        smiles = smiles.replace("\\/", "/")
        mol = Chem.MolFromSmiles(smiles)
        if mol:
            smiles = Chem.MolToSmiles(mol, kekuleSmiles=True)
            splitted_list = list(smiles)
            tokenized_SMILES = re.sub(
                r"\s+(?=[a-z])", "", " ".join(map(str, splitted_list))
            )
            decoded = tokenize_input(tokenized_SMILES, inp_lang, inp_max_length)
            result_predited = self.reloaded_forward(decoded)
            result = detokenize_output(result_predited, targ_lang)
            return result
        else:
            return "Could not generate IUPAC name from invalid SMILES."

    def iupace_to_smiles(self, iupacname: str) -> str:
        """Takes user input splits them into words and generates tokens.
        Tokens are then passed to the model and the model predicted tokens are retrieved.
        The predicted tokens gets detokenized and the final result is returned in a string format.
        Args:
            iupacname (str): user input IUPAC names in string format.
        Returns:
            result (str): The predicted SMILES in string format.
        """

        # Load important pickle files which consists the tokenizers and the maxlength setting
        targ_lang = pickle.load(
            open(self.default_path.as_posix() + "/assets/tokenizer_input.pkl", "rb")
        )
        inp_lang = pickle.load(
            open(self.default_path.as_posix() + "/assets/tokenizer_target.pkl", "rb")
        )
        inp_max_length = pickle.load(
            open(self.default_path.as_posix() + "/assets/max_length_targ.pkl", "rb")
        )

        splitted_list = list(iupacname)
        tokenized_IUPACname = " ".join(map(str, splitted_list))
        decoded = tokenize_input(tokenized_IUPACname, inp_lang, inp_max_length)

        result_predited = self.reloaded_reverse(decoded)
        result = detokenize_output(result_predited, targ_lang)

        return result


# Converts the unicode file to ascii
def unicode_to_ascii(s: str) -> str:
    """Converts a unicode string to an ASCII string

    Args:
        s (str): Takes a string in unicode format.

    Returns:
        str: returns a ASCII formatted string.
    """

    return "".join(
        c for c in unicodedata.normalize("NFD", s) if unicodedata.category(c) != "Mn"
    )


def preprocess_sentence(w: str) -> str:
    """Takes in a sentence, removes white spaces and generates a clean sentence. At the begining of the sentesnce a <start> token will be added
    and at the end an <end> token will be added and the modified sentence will be returned.

    Args:
        w (str): input sentence to be modified.

    Returns:
        str: modified sentence with start and end tokens.
    """
    w = unicode_to_ascii(w.strip())

    # creating a space between a word and the punctuation following it
    # eg: "he is a boy." => "he is a boy ."
    # Reference:- https://stackoverflow.com/questions/3645931/python-padding-punctuation-with-white-spaces-keeping-punctuation
    w = re.sub(r"([?.!,¿])", r" \1 ", w)
    w = re.sub(r'[" "]+', " ", w)

    # replacing everything with space except (a-z, A-Z, ".", "?", "!", ",")
    # w = re.sub(r"[^a-zA-Z?.!,¿]+", " ", w)

    w = w.strip()

    # adding a start and an end token to the sentence
    # so that the model know when to start and stop predicting.
    w = "<start> " + w + " <end>"
    return w


def tokenize_input(input_SMILES: str, inp_lang, inp_max_length: int) -> np.array:
    """This function takes a user input SMILES and tokenizes it
       to feed it to the model.

    Args:
        input_SMILES (string): SMILES string given by the user.
        inp_lang: keras_preprocessing.text.Tokenizer object with input language.
        inp_max_length: maximum number of characters in the input language.

    Returns:
        tokenized_input (np.array): The SMILES get split into meaningful chunks
        and gets converted into meaningful tokens. The tokens are arrays.
    """
    sentence = preprocess_sentence(input_SMILES)
    inputs = [inp_lang.word_index[i] for i in sentence.split(" ")]
    tokenized_input = keras.preprocessing.sequence.pad_sequences(
        [inputs], maxlen=inp_max_length, padding="post"
    )

    return tokenized_input


def detokenize_output(predicted_array: np.array, targ_lang) -> str:
    """This function takes a predited input array and returns
       a IUPAC name by detokenizing the input.

    Args:
        predicted_array (np.array): The predicted_array is returned by the model.
        targ_lang: keras_preprocessing.text.Tokenizer object with target language.

    Returns:
        prediction (str): The predicted array gets detokenized by the tokenizer,
        The unnessary spaces, start and the end tokens will bve removed and
        a proper IUPAC name is returned in a string format.
    """
    outputs = [targ_lang.index_word[i] for i in predicted_array[0].numpy()]
    prediction = (
        " ".join([str(elem) for elem in outputs])
        .replace("<start> ", "")
        .replace(" <end>", "")
        .replace(" ", "")
    )

    return prediction


def create_look_ahead_mask(size):
    mask = 1 - tf.linalg.band_part(tf.ones((size, size)), -1, 0)
    return mask  # (seq_len, seq_len)


def create_padding_mask(seq):
    seq = tf.cast(tf.math.equal(seq, 0), tf.float32)

    # add extra dimensions to add the padding
    # to the attention logits.
    return seq[:, tf.newaxis, tf.newaxis, :]  # (batch_size, 1, 1, seq_len)


def create_masks(inp, tar):
    # Encoder padding mask
    enc_padding_mask = create_padding_mask(inp)

    # Used in the 2nd attention block in the decoder.
    # This padding mask is used to mask the encoder outputs.
    dec_padding_mask = create_padding_mask(inp)

    # Used in the 1st attention block in the decoder.
    # It is used to pad and mask future tokens in the input received by
    # the decoder.
    look_ahead_mask = create_look_ahead_mask(tf.shape(tar)[1])
    dec_target_padding_mask = create_padding_mask(tar)
    combined_mask = tf.maximum(dec_target_padding_mask, look_ahead_mask)

    return enc_padding_mask, combined_mask, dec_padding_mask


# Downloads the model and unzips the file downloaded, if the model is not present on the working directory.
def download_trained_weights(model_url: str, model_path: str, verbose=1):
    """This function downloads the trained models and tokenizers to a default location.
    After downloading the zipped file the function unzips the file automatically.
    If the model exists on the default location this function will not work.

    Args:
        model_url (str): trained model url for downloading.
        model_path (str): model default path to download.

    Returns:
        downloaded model.
    """
    # Download trained models
    if verbose > 0:
        logger.info("Downloading trained model to " + str(model_path))
    model_path = pystow.ensure("STOUT-V2", url=model_url)
    if verbose > 0:
        logger.info("... done downloading trained model!")
    subprocess.run(
        [
            "unzip",
            model_path.as_posix(),
            "-d",
            model_path.parent.as_posix(),
        ]
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pura', 'pura.services']

package_data = \
{'': ['*']}

install_requires = \
['Pint>=0.19.2,<0.20.0',
 'aiohttp>=3.8.1,<4.0.0',
 'lxml>=4.9.0,<5.0.0',
 'nest-asyncio>=1.5.6,<2.0.0',
 'pandas>=1.4.3,<2.0.0',
 'pydantic>=1.9.1,<2.0.0',
 'python-dotenv>=0.20.0,<0.21.0',
 'rdkit-pypi>=2022.3.3,<2023.0.0',
 'tqdm>=4.64.0,<5.0.0']

extras_require = \
{'stout': ['unicodedata2>=14.0.0,<15.0.0',
           'pystow>=0.4.5,<0.5.0',
           'tensorflow-gpu==2.7.2']}

setup_kwargs = {
    'name': 'pura',
    'version': '0.2.3',
    'description': 'Clean chemical data quickly',
    'long_description': 'None',
    'author': 'Kobi Felton',
    'author_email': 'kobi.c.f@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



from pydantic import BaseModel
import pint
from typing import List, TypeVar, Generic


T = TypeVar("T")
# from https://github.com/hgrecco/pint/issues/1166#issuecomment-1116309404
class PintType(Generic[T]):

    Q = pint.Quantity

    def __init__(self, q_check: str):
        self.q_check = q_check

    def __get_validators__(self):
        yield self.validate

    def validate(self, v):
        q = self.Q(v)
        assert q.check(self.q_check), f"Dimensionality must be {self.q_check}"
        return q


Mass = PintType("[mass]")
Amount = PintType("[substance]")
Volume = PintType("[volume]")
Time = PintType("[time]")
Temperature = PintType("[temperature]")
